__all__ = (
    'SqlException',
    'QueryException',
)


class SqlException(Exception):
    """ Sql building/parsing error. """
    __slots__ = ()

    def __init__(self, message: str):
        self.args = (message, )

    @property
    def message(self):
        return self.args[0]


class QueryException(Exception):
    """ Query to database error. """
    __slots__ = ()

    def __init__(self, sql: str, message: str):
        self.args = (message, sql)

    @property
    def message(self):
        return self.args[0]

    @property
    def sql(self):
        return self.args[1]



from passql.exceptions import QueryException
from passql.interfaces import *
from passql.models import DbEntity
from passql.sql import Sql, SqlMaker
from typing import Union, Any, List, Type, TypeVar, Optional, Coroutine

__all__ = (
    'DbConnection',
)

TEntity = TypeVar('TEntity', bound=DbEntity)
T = TypeVar('T')


class DbConnection:
    __slots__ = ('_connection', '_sql_maker')

    def __init__(self, origin_connection: IOriginConnection, sql_maker: SqlMaker):
        self._connection = origin_connection
        self._sql_maker = sql_maker

    @property
    def origin_connection(self) -> IOriginConnection:
        """ Connection received in constructor. """
        return self._connection

    def transaction(self) -> IOriginTransaction:
        """ Get transaction object. """
        return self._connection.transaction()

    def execute(self,
                sql: Union[Sql, str],
                params: Any = None,
                timeout: int = None) -> Coroutine[Any, Any, None]:
        """ Execute sql.

        :param sql: Query string or prepared Sql object.
        :param params: Parameters to parse to sql.
        :param timeout: Time limit for the operation.
        """

        if type(sql) is str and params is not None:
            sql = self._sql_maker(sql).format(params)
        else:
            sql = sql.format(params)

        try:
            return self._connection.execute(sql, timeout=timeout)
        except Exception as e:
            raise QueryException(sql, str(e))

    async def query_scalar(self,
                           t: Type[T],
                           sql: Union[Sql, str],
                           params: Any = None,
                           timeout: int = None) -> Optional[T]:
        """ Query single value.

        :param t: Return scalar type, only for generic.
        :param sql: Query string or prepared Sql object.
        :param params: Parameters to parse to sql.
        :param timeout: Time limit for the operation.
        :return: Value or None.
        """

        if type(sql) is str and params is not None:
            sql = self._sql_maker(sql).format(params)
        else:
            sql = sql.format(params)

        try:
            return await self._connection.fetchval(sql, timeout=timeout)
        except Exception as e:
            raise QueryException(sql, str(e))

    async def query_list(self,
                         t: Type[TEntity],
                         sql: Union[Sql, str],
                         params: Any = None,
                         timeout: int = None) -> List[TEntity]:
        """ Query list of entities.

        :param t: Return entity type.
        :param sql: Query string or prepared Sql object.
        :param params: Parameters to parse to sql.
        :param timeout: Time limit for the operation.
        :return: List of parsed entities.
        """

        if type(sql) is str and params is not None:
            sql = self._sql_maker(sql).format(params)
        else:
            sql = sql.format(params)

        try:
            result = await self._connection.fetch(sql, timeout=timeout)
        except Exception as e:
            raise QueryException(sql, str(e))

        for i in range(len(result)):
            result[i] = t.from_record(result[i])

        # noinspection PyTypeChecker
        return result

    async def query_values_list(self,
                                t: Type[T],
                                sql: Union[Sql, str],
                                params: Any = None,
                                timeout: int = None) -> List[T]:
        """ Query list of entities.

        :param t: Return record/scalar type, only for generic.
        :param sql: Query string or prepared Sql object.
        :param params: Parameters to parse to sql.
        :param timeout: Time limit for the operation.
        :return: List of objects.
        """

        if type(sql) is str and params is not None:
            sql = self._sql_maker(sql).format(params)
        else:
            sql = sql.format(params)

        try:
            result = await self._connection.fetch(sql, timeout=timeout)
        except Exception as e:
            raise QueryException(sql, str(e))

        if result:
            if len(result[0]) > 1:
                for i in range(len(result)):
                    # noinspection PyTypeChecker
                    result[i] = tuple(result[i].values())
            else:
                for i in range(len(result)):
                    # noinspection PyTypeChecker
                    result[i] = next(result[i].values())

        return result

    async def query_first(self,
                          t: Type[TEntity],
                          sql: Union[Sql, str],
                          params: Any = None,
                          timeout: int = None) -> Optional[TEntity]:
        """ Query first entity.

        :param t: Return entity type.
        :param sql: Query string or prepared Sql object.
        :param params: Parameters to parse to sql.
        :param timeout: Time limit for the operation.
        :return: Entity parsed from first result row.
        """

        if type(sql) is str and params is not None:
            sql = self._sql_maker(sql).format(params)
        else:
            sql = sql.format(params)

        try:
            result = await self._connection.fetchrow(sql, timeout=timeout)
        except Exception as e:
            raise QueryException(sql, str(e))

        return None if result is None else t.from_record(result)

    async def close(self):
        """ Close connection. """
        await self._connection.close()

    def is_closed(self) -> bool:
        """ Is connection closed. """
        return self._connection.is_closed()



from passql.sql import SqlConverter
import datetime

__all__ = (
    'SqlDefaultConverters',
)


class SqlDefaultConverters:
    COMMON = SqlConverter({
        int: lambda val, _: str(val),
        float: lambda val, _: str(val),
        bool: lambda val, _: "TRUE" if val else "FALSE",
        type(None): lambda val, _: "NULL",
    })

    POSTGRES = COMMON + SqlConverter({
        str: lambda val, converter: "'" + val.replace("'", "") + "'::text",
        tuple: lambda val, converter: ','.join(converter(v) for v in val),
        range: lambda val, converter: ','.join(converter(v) for v in val),
        list: lambda val, converter: "'{" + ','.join(converter(v) for v in val) + "}'",
        set: lambda val, converter: "'{" + ','.join(converter(v) for v in val) + "}'",
        datetime.datetime: lambda val, converter: f"'{val}'::timestamp",
        datetime.date: lambda val, converter: f"'{val}'::date",
    })



from passql.interfaces import *
from typing import Set, Optional
from asyncpg import Record
import inspect

__all__ = (
    'DbEntity',
    'register_entities',
)


class DbEntity:
    __rec: Optional[Record]
    __fields: Set[str]

    def __init__(self):
        self.__rec = None

    @classmethod
    def _register_as_serializable(cls):
        def add_property(name: str):
            def get_value(self):
                if name in self.__dict__:
                    return self.__dict__[name]
                if self.__rec is not None and \
                        name in cls.__fields and name in self.__rec:
                    return self.__rec[name]
                raise AttributeError(f"{cls.__name__}.{name} field is not set!")

            def set_value(self, value):
                self.__dict__[name] = value

            setattr(cls, name, property(get_value, set_value))

        cls.__fields = set()
        for field_name in cls.__annotations__:
            add_property(field_name)
            cls.__fields.add(field_name)

    @classmethod
    def from_record(cls, rec: IDbRecord):
        obj = cls()
        obj.__rec = rec
        return obj


def register_entities():
    stack = inspect.stack()

    for obj in stack[1][0].f_locals.values():
        if type(obj) is type:
            if DbEntity in obj.__bases__:
                # noinspection PyProtectedMember
                obj._register_as_serializable()



from typing import Protocol, Any, List, Iterator

__all__ = (
    'IDbRecord',
    'IOriginTransaction',
    'IOriginConnection',
)


class IDbRecord(Protocol):
    def __contains__(self, field_name) -> bool:
        ...

    def __getitem__(self, field_name) -> Any:
        ...

    def __len__(self) -> int:
        ...

    def keys(self) -> Iterator[str]:
        ...

    def values(self) -> Iterator[Any]:
        ...


class IOriginTransaction(Protocol):
    async def start(self):
        ...

    async def commit(self):
        ...

    async def rollback(self):
        ...

    async def __aenter__(self):
        ...

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        ...


class IOriginConnection(Protocol):
    async def fetch(self, sql: str, timeout=None) -> List[IDbRecord]:
        ...

    async def fetchrow(self, sql: str, timeout=None) -> IDbRecord:
        ...

    async def execute(self, sql: str, timeout=None):
        ...

    async def fetchval(self, sql: str, timeout=None) -> Any:
        ...

    async def close(self):
        ...

    def is_closed(self) -> bool:
        ...

    def transaction(self) -> IOriginTransaction:
        ...



""" python setup.py sdist
    python publish.py
"""

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='passql',
      version='0.2.5',

      description='Super light ORM',
      long_description=long_description,
      long_description_content_type='text/markdown',

      author='Vladislav Mironov',
      author_email='hidden120@mail.ru',

      python_requires='>=3.6',
      classifiers=[
          'Programming Language :: Python :: 3.9',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
      ],

      packages=['passql'],
      zip_safe=False)



from passql.exceptions import SqlException
from typing_extensions import TypeAlias
from typing import Dict, Type, Any, Callable, Optional
import re

__all__ = (
    'Sql',
    'SqlMaker',
    'SqlConverter',
)

ValueToSqlConverter: TypeAlias = Callable[[Any, 'SqlConverter'], str]


class Sql:
    __slots__ = ('_converter', '_strings', '_params')

    __PRM_PATTERN: re.Pattern[str] = re.compile(r"@[\w]+")

    def __init__(self, converter: 'SqlConverter', string: str):
        self._converter = converter
        self._strings = []
        self._params = []

        string = string.strip()
        last = 0
        for m in re.finditer(Sql.__PRM_PATTERN, string):
            start = m.start(0)

            if start - 2 > 0:
                if string[start - 1] == '\\':
                    if string[start - 2] != '\\':
                        continue
            elif start - 1 > 0:
                if string[start - 1] == '\\':
                    continue

            s = string[last:start]
            if s:
                self._strings.append(s)

            last = m.end(0)
            self._params.append(string[start + 1:last])

        s = string[last:]
        if s:
            self._strings.append(s)

    def __repr__(self):
        if self._strings:
            short_preview = self._strings[0][:10] + \
                            ('..., ' if len(self._strings) > 1 or len(self._strings[0]) > 10 else ', ')
        else:
            short_preview = ''
        params = f"{len(self._params)} param" + ('' if self._params == 1 else 's')
        return f"<Sql: {short_preview}{params}>"

    def __str__(self):
        if self._params:
            raise SqlException(f"Cannot convert {self.__class__.__name__} to string, parameters required!")

        return ''.join(self._strings)

    def format(self, obj: Optional) -> str:
        """
        Parse parameters from obj to sql template.

        :param obj: Any object or dictionary with required fields/keys.
        """
        result = []
        prm_length = len(self._params)

        if type(obj) is dict:
            for i in range(len(self._strings)):
                result.append(self._strings[i])
                if i < prm_length:
                    value = obj[self._params[i]]
                    result.append(value.format(obj) if type(value) is Sql else self._converter(value))
        else:
            for i in range(len(self._strings)):
                result.append(self._strings[i])
                if i < prm_length:
                    value = getattr(obj, self._params[i])
                    result.append(value.format(obj) if type(value) is Sql else self._converter(value))

        return ''.join(result)


class SqlMaker:
    __slots__ = ('_converter', )

    def __init__(self, converter: 'SqlConverter'):
        self._converter = converter

    def __call__(self, string: str) -> Sql:
        return Sql(self._converter, string)


class SqlConverter:
    __slots__ = ('_type_to_converter', )

    def __init__(self, type_to_converter_map: Dict[Type, ValueToSqlConverter]):
        self._type_to_converter = {
            Sql: lambda val, _: str(val),
        }

        for t in type_to_converter_map:
            self._type_to_converter[t] = type_to_converter_map[t]

    def __call__(self, value: Any) -> str:
        converter = self._type_to_converter.get(type(value))
        if converter is None:
            raise SqlException(f"Type of mapping object '{value}' has no any convertor!")

        return converter(value, self)

    def __add__(self, other: 'SqlConverter') -> 'SqlConverter':
        type_to_converters = {}

        for t, converter in self._type_to_converter.items():
            type_to_converters[t] = converter
        for t, converter in other._type_to_converter.items():
            type_to_converters[t] = converter

        return SqlConverter(type_to_converters)
